(    Title:  Forth 200x REQUIRED reference implementation module
      File:  required.fs
   Adapter:  David N. Williams
   License:  Public Domain
   Revised:  February 14, 2012

This file makes the Forth 200x reference implementation of
REQUIRED and REQUIRE into an anonymous module:

  http://www.forth200x.org/required.html
)

\ This file is in the public domain. NO WARRANTY.

\ This implementation does not implement the requirements wrt MARKER
\ and FORGET (REQUIRED only includes each file once, whether a marker
\ was executed or not), so it is not a correct implementation on
\ systems that support these words.

BEGIN-MODULE

\ The program uses the following words
\ from CORE :
\  environment? 0= : swap >r dup 2dup r> rot move ; cells r@ @ over ! cell+ 
\  2! BEGIN WHILE 2@ IF drop 2drop EXIT THEN REPEAT ELSE Variable 
\ from CORE-EXT :
\  2>r 2r@ 2r> true 
\ from BLOCK-EXT :
\  \ 
\ from EXCEPTION :
\  throw 
\ from FILE :
\  S" ( included 
\ from MEMORY :
\  allocate 
\ from STRING :
\  compare 
\ from TOOLS-EXT :
\  [IF] [THEN]

PRIVATE:

\ we use a linked list of names

[UNDEFINED] PARSE-NAME [IF]
\ http://www.forth200x.org/reference-implementations/parse-name.fs

: isspace?    ( c -- f )  bl 1+ u< ;
: isnotspace? ( c -- f )  isspace? 0= ;

: xt-skip   ( addr1 n1 xt -- addr2 n2 ) \ gforth
    \ skip all characters satisfying xt ( c -- f )
    >r
    BEGIN
	dup
    WHILE
	over c@ r@ execute
    WHILE
	1 /string
    REPEAT  THEN
    r> drop ;

: parse-name ( "name" -- c-addr u )
    source >in @ /string
    ['] isspace? xt-skip over >r
    ['] isnotspace? xt-skip ( end-word restlen r: start-word )
    2dup 1 min + source drop - >in !
    drop r> tuck - ;

[THEN] \ PARSE-NAME

: save-mem	( addr1 u -- addr2 u ) \ gforth
    \ copy a memory block into a newly allocated region in the heap
    swap >r
    dup allocate throw
    swap 2dup r> rot rot move ;

: name-add ( addr u listp -- )
    >r save-mem ( addr1 u )
    3 cells allocate throw \ allocate list node
    r@ @ over ! \ set next pointer
    dup r> ! \ store current node in list var
    cell+ 2! ;
    
: name-present? ( addr u list -- f )
    rot rot 2>r begin ( list R: addr u )
	dup
    while
	dup cell+ 2@ 2r@ compare 0= if
	    drop 2r> 2drop true EXIT
	then
	@
    repeat
    ( drop 0 ) 2r> 2drop ;

: name-join ( addr u list -- )
    >r 2dup r@ @ name-present? if
	r> drop 2drop
    else
	r> name-add
    then ;

variable included-names 0 included-names !

: included ( i*x addr u -- j*x )
    2dup included-names name-join
    included ;

: include ( i*x "name" -- j*x )
    parse-name included ;

PUBLIC:

: required ( i*x addr u -- i*x )
    2dup included-names @ name-present? 0= if
	included
    else
	2drop
    then ;

: require ( i*x "name" -- i*x )
    parse-name required ;

END-MODULE
